/* ************************************************************************** */
/*                                       _____           _                    */
/*  parse/lang.c                        |_   _|__  _   _| |__   ___ _   _     */
/*  | Project: libg1m                     | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/12/01 13:33:45                               |___/     */
/*                                                                            */
/* ************************************************************************** */
#include <libg1m/internals.h>

/**
 *	g1m_parse_lang:
 *	Parse fx language files.
 *
 *	@arg	handle		the libg1m handle.
 *	@arg	stream		the stream to parse from.
 *	@return				the error code (0 if ok).
 */

int g1m_parse_lang(g1m_t *handle, FILE *stream,
	struct standard_header *std)
{
	(void)std;
	/* set handle type */
	handle->type = g1m_type_lang;
	handle->messages = NULL;

	/* read the subheader */
	DREAD(hd, g1l_subheader)

	/* correct the endianness */
	uint_fast32_t num = be16toh(hd.message_count) + 1;

	/* log */
	log_info("%lu messages to read", num);

	/* beat the best, read the rest! */
	size_t data_size = std->filesize - sizeof(struct standard_header)
		- sizeof(struct g1l_subheader);
	uint8_t data[data_size];
	READ(data, data_size)

	/* allocate tab */
	handle->messages = malloc(sizeof(char*) * num);
	handle->count = 0;
	handle->_size = num;
	if (!handle->messages)
		return (g1m_error_alloc);

	/* get the offset table */
	uint16_t *offsets = (uint16_t*)data;
	char *messages = (char*)(offsets + num);

	/* read messages */
	for (uint_fast32_t i = 0; i < num; i++) {
		handle->messages[i] = NULL;
		if (offsets[i] == (uint16_t)-1) {
			log_info("[#%lu] -", i);
			handle->count++;
			continue ;
		}

		/* correct offset */
		offsets[i] = be16toh(offsets[i]);

		/* log */
		log_info("[#%lu] '%s' (0x%hu)", i, &messages[offsets[i]], offsets[i]);

		/* store */
		handle->messages[i] = strdup(&messages[offsets[i]]);
		if (!handle->messages[i]) goto fail;
		handle->count++;
	}

	/* no error */
	return (0);

	/* omg fail! */
fail:
	for (int i = 0; i < handle->count; i++)
		free(handle->messages[i]);
	free(handle->messages);
	handle->messages = NULL;
	return (g1m_error_alloc);
}

/**
 *	g1m_parse_lang_cg:
 *	Parse fx-CG language files.
 *
 *	@arg	handle		the libg1m handle.
 *	@arg	stream		the stream to parse from.
 *	@arg	std			the standard header.
 *	@return				the error code (0 if ok).
 */

int g1m_parse_lang_cg(g1m_t *handle, FILE *stream,
	struct standard_header *std)
{
	(void)handle;
	/* set handle type */
	handle->type = g1m_type_lang;
	handle->is_cg = 1;
	handle->messages = NULL;

	/* read the subheader */
	DREAD(hd, g3l_subheader)

	/* correct the endianness */
	hd.checksum = be32toh(hd.checksum);
	hd.message_zone_size = be32toh(hd.message_zone_size);

	/* log */
	log_info("internal name is '%.8s'", hd.internal_name);
	log_info("language name is '%.16s'", hd.language_name);
	log_info("language salutation is '%.16s'", hd.language_salutation);
	log_info("version is '%.10s'", hd.version);
	log_info("datetime is '%.14s'", hd.datetime);
	log_info("g3l filename is '%.16s'", hd.g3l_filename);

	/* read the languages header */
	DREAD(lhd, g3l_lang_header)
	logm_info(&lhd, sizeof(struct g3l_lang_header));

	/* correct endianness */
	lhd.num = be32toh(lhd.num);

	/* log */
	log_info("there are %d messages", lhd.num);

	/* read the entire message zone */
	size_t message_zone_size = hd.message_zone_size
		- sizeof(struct g3l_lang_header);
	uint8_t message_zone[message_zone_size];
	READ(message_zone, message_zone_size)

	/* calculate checksum and check */
	uint32_t checksum =
		sizeof(struct standard_header) * 255 - g1m_checksum32(std, sizeof(struct standard_header), 0) +
		g1m_checksum32(&hd.magic, sizeof(struct g3l_subheader) - 4, 0) +
		g1m_checksum32(&lhd, sizeof(struct g3l_lang_header), 0) +
		g1m_checksum32(message_zone, message_zone_size, 0);
	log_info("header checksum is %u, calc checksum is %u",
		hd.checksum, checksum);
	if (hd.checksum && hd.checksum != checksum) {
		log_fatal("header checksum was invalid");
		return (g1m_error_magic);
	}

	/* parse footer */
	DREAD(footer, g3l_footer)
	footer.checksum = be32toh(footer.checksum);
	if (footer.checksum != hd.checksum) {
		log_fatal("footer checksum was invalid");
		return (g1m_error_magic);
	}

	/* get parts */
	const uint_fast32_t num = lhd.num + 1;
	uint32_t *offsets = (uint32_t*)message_zone;
	char *messages = (char*)message_zone + num * sizeof(uint32_t);

	/* prepare tab */
	handle->messages = malloc(sizeof(char*) * num);
	handle->count = 0;
	handle->_size = num;
	if (!handle->messages)
		return (g1m_error_alloc);

	/* read messages */
	for (uint_fast32_t i = 0; i < num; i++) {
		/* check if offset is valid */
		handle->messages[i] = NULL;
		if (offsets[i] == (uint32_t)-1) {
			log_info("[#%ld] -", i);
			handle->count++;
			continue ;
		}

		/* correct offset */
		offsets[i] = be32toh(offsets[i]);

		/* log */
		log_info("[#%ld] '%s'", i, &messages[offsets[i]]);

		/* store message */
		handle->messages[i] = strdup(&messages[offsets[i]]);
		if (!handle->messages[i]) goto fail;
		handle->count++;
	}

	/* no error */
	return (0);

	/* omg fail! */
fail:
	for (int i = 0; i < handle->count; i++)
		free(handle->messages[i]);
	free(handle->messages);
	handle->messages = NULL;
	return (g1m_error_alloc);
}
