/* ************************************************************************** */
/*                                       _____           _                    */
/*  parse/main.c                        |_   _|__  _   _| |__   ___ _   _     */
/*  | Project: libg1m                     | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/12/01 13:33:45                               |___/     */
/*                                                                            */
/* ************************************************************************** */
#include <libg1m/internals.h>

/**
 *	g1m_parse:
 *	Parse a G1M file.
 *
 *	Read the standard header, correct endianness, check magic numbers,
 *	then read subparts according to the G1M type.
 *
 *	@arg	handle		the handle.
 *	@arg	stream		the stream to parse from.
 *	@return				the error code (0 if ok).
 */

int g1m_parse(g1m_t *handle, FILE *stream)
{
	/* initialize the handle */
	bzero(handle, sizeof(g1m_t));

	/* get the standard header */
	DREAD(hd, standard_header)

	/* reverse */
	uint8_t *u = (uint8_t*)&hd;
	for (size_t i = 0; i < sizeof(struct standard_header); i++) u[i] = ~u[i];

	/* print header */
	log_info("Raw inverted standard header is:");
	logm_info(&hd, sizeof(struct standard_header));

	/* correct standard header endianess */
	hd.filesize = be32toh(hd.filesize);
	hd.number = be16toh(hd.number);

	/* get type */
	int (*read_func)(g1m_t*, FILE*, struct standard_header*);
	const char *type_string; int check_one, check_two;
	if (g1m_get_type_info(hd.main_id, hd.subtype, &type_string,
		&check_one, &check_two, &read_func))
		return (g1m_error_magic);

	/* check standard header magics */
	if (check_one && hd.control != ((hd.filesize + 0x41) & 0xff)) {
		log_info("First control byte isn't right.");
		return (g1m_error_magic);
	} else if (check_two && hd.control2 != ((hd.filesize + 0xb8) & 0xff)) {
		log_info("Second control byte isn't right.");
		return (g1m_error_magic);
	}

	/* log some data */
	log_info("Standard Header type is '%s'.", type_string);
	log_info("Standard Header filesize is %uo.", hd.filesize);
	log_info("Standard Header num is %d.", hd.number);

	/* subparse. */
	handle->type = 0;
	handle->is_cg = 0;
	return ((*read_func)(handle, stream, &hd));
}
