/* ************************************************************************** */
/*                                       _____           _                    */
/*  user/free.c                         |_   _|__  _   _| |__   ___ _   _     */
/*  | Project: libg1m                     | |/ _ \| | | | '_ \ / _ \ | | |    */
/*                                        | | (_) | |_| | | | |  __/ |_| |    */
/*  By: thomas <thomas@touhey.fr>         |_|\___/ \__,_|_| |_|\___|\__, |.fr */
/*  Last updated: 2016/11/21 09:13:10                               |___/     */
/*                                                                            */
/* ************************************************************************** */
#include <libg1m/internals.h>
#include <stdlib.h>

/**
 *	g1m_free_line_content:
 *	Free an e-activity line content.
 *
 *	@arg	handle		the line handle.
 */

void g1m_free_line_content(g1m_line_t *line)
{
	if (line->type & (g1m_linetype_title | g1m_linetype_text
	  | g1m_linetype_picture))
		free(line->content);

	if (line->type & g1m_linetype_eact) {
		for (int i = 0; i < line->count; i++) {
			g1m_free_line_content(line->lines[i]);
			free(line->lines[i]);
		}
		free(line->lines);
	}
}

/**
 *	g1m_free_mcsfile_content:
 *	Free an MCS file handle content.
 *
 *	@arg	handle		the handle which has the content to free.
 */

void g1m_free_mcsfile_content(g1m_mcsfile_t *handle)
{
	/* free content for unknown files */
	if (!handle->type)
		free(handle->content);

	/* free cells */
	if (handle->type & (g1m_mcstype_mat | g1m_mcstype_vct | g1m_mcstype_list
	  | g1m_mcstype_spreadsheet)
	 && handle->columns && handle->rows) {
		free(handle->cells[0]);
		free(handle->cells);
	}

	/* free the images */
	if (handle->type & (g1m_mcstype_pict | g1m_mcstype_capt)) {
		free(handle->image);
		if (handle->type & g1m_mcstype_capt)
			free(handle->second_image);
	}

	/* free the content */
	if (handle->type & g1m_mcstype_program)
		free(handle->content);
}

/**
 *	g1m_free_mcsfile:
 *	Free an MCS file handle.
 *
 *	@arg	handle		the handle to free.
 */

void g1m_free_mcsfile(g1m_mcsfile_t *handle)
{
	g1m_free_mcsfile_content(handle);
	free(handle);
}

/**
 *	g1m_free_mcs:
 *	Free all of the MCS.
 *
 *	@arg	handle		the handle to close.
 */

void g1m_free_mcs(g1m_t *handle)
{
	/* check if mcs */
	if (!handle->files)
		return ;

	/* foreach file in mcs */
	g1m_mcsfile_t **files = handle->files;
	int file_count = handle->count;
	for (int i = 0; i < file_count; i++) {
		/* free the file if exists */
		if (files[i]) g1m_free_mcsfile(files[i]);
	}
	free(files);
	handle->files = NULL;
}

/**
 *	g1m_free:
 *	Free handle data.
 *
 *	@arg	handle		the handle to close.
 */

void g1m_free(g1m_t *handle)
{
	if (!handle) return ;

	/* addin time! */
	if (handle->type & g1m_type_addin)
		free(handle->pixels);

	/* mcs time! */
	if (handle->type & g1m_type_mcs)
		g1m_free_mcs(handle);

	/* messages time! */
	if (handle->type & g1m_type_lang
	 && handle->messages) {
		for (int i = 0; i < handle->count; i++)
			free(handle->messages[i]);
		free(handle->messages);
	}

	/* function keys time! */
	if (handle->type & g1m_type_fkey
	 && handle->fkeys) {
		for (int i = 0; i < handle->count; i++)
			free(handle->fkeys[i]);
		free(handle->fkeys);
	}

	/* picture time! */
	if (handle->type & g1m_type_picture)
		free(handle->pixels);

	/* e-activities time! */
	if (handle->type & g1m_type_eact)
		g1m_free_line_content(handle->line);

	/* free the handle */
	free(handle);
}
